###############################################################################
#  b_state.py: Module for PATTERN MATCH - base state class
#  Copyright (C) 2010 Brno University of Technology, ANT @ FIT
#  Author(s): Vlastimil Kosar <ikosar@fit.vutbr.cz>
###############################################################################
#
#  LICENSE TERMS
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#  3. All advertising materials mentioning features or use of this software
#     or firmware must display the following acknowledgement:
#
#       This product includes software developed by the University of
#       Technology, Faculty of Information Technology, Brno and its
#       contributors.
#
#  4. Neither the name of the Company nor the names of its contributors
#     may be used to endorse or promote products derived from this
#     software without specific prior written permission.
#
#  This software or firmware is provided ``as is'', and any express or implied
#  warranties, including, but not limited to, the implied warranties of
#  merchantability and fitness for a particular purpose are disclaimed.
#  In no event shall the company or contributors be liable for any
#  direct, indirect, incidental, special, exemplary, or consequential
#  damages (including, but not limited to, procurement of substitute
#  goods or services; loss of use, data, or profits; or business
#  interruption) however caused and on any theory of liability, whether
#  in contract, strict liability, or tort (including negligence or
#  otherwise) arising in any way out of the use of this software, even
#  if advised of the possibility of such damage.
#
#  $Id$


class b_State:
    """A base class for state representation."""
    def __init__(self, mid = 0, rnum = 0):
        self._id    = mid;   # State identification value
        self._rnum  = rnum;  # Regular expression numeber (final state) 
 
    def get_text(self):
        """Return text description for graph representation."""
        return str(self._id)

    def get_id(self):
        """Return state identification number."""
        return self._id

    def is_final(self):
        """Return true if the state is a final state."""
        return self._rnum!=-1

    def get_regexp_number(self):
        """Return index of regular expression, which corresponds to the
        final state. If 0 value is returned the state is not final and
        do not represent any regular expression."""
        return self._rnum;
    
    def __str__(self):
        """ Returns state identification value as string."""
        return str(self._id)

    def __repr__(self):
        """ Returns representation of state identification value and regular expression numeber."""
        return "<" + str(self._id) + ", " + str(self._rnum) + ">"
        
        
###############################################################################
# End of File b_state.py                                                      #
###############################################################################